@@ -1,5 +1,3 @@
-import SHA
-
 function iterate_headers(
     callback::Function,
     tar::IO;
@@ -32,7 +30,13 @@ function iterate_headers(
     @label loop
         hdr === nothing && break
         strict && check_header(hdr)
-        callback(hdr)
+        if hasmethod(callback, Tuple{Header, Vector{Pair{Symbol, String}}})
+            callback(hdr, dump_header(buf))
+        elseif hasmethod(callback, Tuple{Header, Vector{UInt8}})
+            callback(hdr, buf[1:512])
+        else
+            callback(hdr)
+        end
         if !skeleton || hdr.type in (:g, :x) ||
             hdr.path == "././@LongLink" && hdr.type in (:L, :K)
             skip_data(tar, hdr.size)
@@ -40,6 +44,24 @@ function iterate_headers(
     end
 end
 
+# follow chains of symlinks
+follow_symlink_chain(seen::Vector, what::Any, paths) =
+    what == :symlink ? what : seen[end]
+follow_symlink_chain(seen::Vector, what::String, paths) =
+    what in seen ? :symlink :
+        follow_symlink_chain(push!(seen, what), paths[what], paths)
+
+# our `cp` doesn't copy ACL properties, so manually set them via `chmod`
+function copy_mode(src::String, dst::String)
+    chmod(dst, filemode(src))
+    isdir(dst) || return
+    for name in readdir(dst)
+        sub_src = joinpath(src, name)
+        sub_dst = joinpath(dst, name)
+        copy_mode(sub_src, sub_dst)
+    end
+end
+
 function extract_tarball(
     predicate::Function,
     tar::IO,
@@ -47,17 +69,26 @@ function extract_tarball(
     buf::Vector{UInt8} = Vector{UInt8}(undef, DEFAULT_BUFFER_SIZE),
     skeleton::IO = devnull,
     copy_symlinks::Bool = false,
+    set_permissions::Bool = true,
 )
+    root = normpath(root)
     paths = read_tarball(predicate, tar; buf=buf, skeleton=skeleton) do hdr, parts
+        Sys.iswindows() && check_windows_path(hdr.path, parts)
         # get the file system version of the path
-        sys_path = reduce(joinpath, init=root, parts)
-        # ensure dirname(sys_path) is a directory
+        sys_path = isempty(parts) ? "." : reduce(joinpath, parts)
+        isabspath(sys_path) &&
+            error("attempt to extract absolute path at $(repr(sys_path)) from $(repr(hdr.path))")
+        sys_path = sys_path == "." ? root : normpath(root, sys_path)
+        startswith(sys_path, root) ||
+            error("attempt to extract relative path outside of root at $(repr(sys_path)) from $(repr(hdr.path))")
+        src_path = joinpath(root, hdr.link)
         dir = dirname(sys_path)
         st = stat(dir)
+        # ensure dirname(sys_path) is a directory
         if !isdir(st)
             ispath(st) && rm(dir, force=true, recursive=true)
             mkpath(dir)
-        else
+        elseif hdr.type != :hardlink || src_path != sys_path
             st = lstat(sys_path)
             hdr.type == :directory && isdir(st) && return # from callback
             ispath(st) && rm(sys_path, force=true, recursive=true)
@@ -67,46 +98,48 @@ function extract_tarball(
             mkdir(sys_path)
         elseif hdr.type == :symlink
             copy_symlinks || symlink(hdr.link, sys_path)
+        elseif hdr.type == :hardlink
+            src_path != sys_path && cp(src_path, sys_path)
         elseif hdr.type == :file
             read_data(tar, sys_path, size=hdr.size, buf=buf)
-            mode = hdr.mode & filemode(sys_path)
-            if 0o100 & hdr.mode == 0
-                # turn off all execute bits
-                mode &= 0o666
-            else
-                # copy read bits to execute bits with
-                # at least the user execute bit on
-                mode |= 0o100 | (mode & 0o444) >> 2
-            end
-            chmod(sys_path, mode)
         else # should already be caught by check_header
             error("unsupported tarball entry type: $(hdr.type)")
         end
+        # apply tarball permissions
+        if set_permissions && hdr.type in (:file, :hardlink)
+            exec = 0o100 & hdr.mode != 0
+            tar_mode = exec ? 0o755 : 0o644
+            sys_mode = filemode(sys_path)
+            if exec
+                # copy read bits to execute bits with
+                # at least the user execute bit on
+                sys_mode |= 0o100 | (sys_mode & 0o444) >> 2
+                # TODO: would be better to have the system
+                # create an executable with default mode but
+                # we don't have a way to do that afaik
+            end
+            chmod(sys_path, tar_mode & sys_mode)
+        end
     end
     copy_symlinks || return
 
     # resolve the internal targets of symlinks
     for (path, what) in paths
-        what isa AbstractString || continue
+        what isa String || continue
         target = link_target(paths, path, what)
         paths[path] = something(target, :symlink)
     end
 
-    # follow chains of symlinks
-    follow(seen::Vector, what::Symbol) =
-        what == :symlink ? what : seen[end]
-    follow(seen::Vector, what::String) =
-        what in seen ? :symlink : follow(push!(seen, what), paths[what])
     for (path, what) in paths
         what isa AbstractString || continue
-        paths[path] = follow([path], what)
+        paths[path] = follow_symlink_chain([path], what, paths)
     end
 
     # copies that need to be made
     copies = Pair{String,String}[]
     for (path, what) in paths
         what isa AbstractString || continue
-        push!(copies, path => what)
+        push!(copies, path => String(what)::String)
     end
     sort!(copies, by=last)
 
@@ -125,17 +158,7 @@ function extract_tarball(
                 src = reduce(joinpath, init=root, split(what, '/'))
                 dst = reduce(joinpath, init=root, split(path, '/'))
                 cp(src, dst)
-                if Sys.iswindows()
-                    # our `cp` doesn't copy ACL properties, so manually set them via `chmod`
-                    function copy_mode(src::String, dst::String)
-                        chmod(dst, filemode(src))
-                        isdir(dst) || return
-                        for name in readdir(dst)
-                            sub_src = joinpath(src, name)
-                            sub_dst = joinpath(dst, name)
-                            copy_mode(sub_src, sub_dst)
-                        end
-                    end
+                if set_permissions && Sys.iswindows()
                     copy_mode(src, dst)
                 end
             end
@@ -145,7 +168,7 @@ end
 
 # resolve symlink target or nothing if not valid
 function link_target(
-    paths::Dict{String,Union{String,Symbol}},
+    paths::Dict{String},
     path::AbstractString,
     link::AbstractString,
 )
@@ -167,7 +190,9 @@ function link_target(
             isempty(link_parts) && break
             what = paths[prefix]
             if what isa AbstractString
+                what = convert(String, what)::String
                 prefix = link_target(paths, prefix, what)
+                prefix === nothing && return
                 path_parts = split(prefix, '/')
             end
         end
@@ -175,7 +200,7 @@ function link_target(
     isempty(path_parts) && return
     target = join(path_parts, '/')
     # if link ends in `/` or `.` target must be a directory
-    part in ("", ".") && paths[target] != :directory && return
+    part in ("", ".") && paths[target] !== :directory && return
     # can't copy a circular link to a prefix of itself
     (path == target || startswith(path, "$target/")) && return
     return target
@@ -184,10 +209,10 @@ end
 function git_tree_hash(
     predicate::Function,
     tar::IO,
-    HashType::DataType,
+    ::Type{HashType},
     skip_empty::Bool;
     buf::Vector{UInt8} = Vector{UInt8}(undef, DEFAULT_BUFFER_SIZE),
-)
+) where HashType <: SHA.SHA_CTX
     # build tree with leaves for files and symlinks
     tree = Dict{String,Any}()
     read_tarball(predicate, tar; buf=buf) do hdr, parts
@@ -206,12 +231,18 @@ function git_tree_hash(
                 node[name] = Dict{String,Any}()
             end
             return
-        end
-        if hdr.type == :symlink
+        elseif hdr.type == :symlink
             mode = "120000"
             hash = git_object_hash("blob", HashType) do io
                 write(io, hdr.link)
             end
+        elseif hdr.type == :hardlink
+            mode = iszero(hdr.mode & 0o100) ? "100644" : "100755"
+            node′ = tree
+            for part in split(hdr.link, '/')
+                node′ = node′[part]
+            end
+            hash = node′[2] # hash of linked file
         elseif hdr.type == :file
             mode = iszero(hdr.mode & 0o100) ? "100644" : "100755"
             hash = git_file_hash(tar, hdr.size, HashType, buf=buf)
@@ -253,8 +284,8 @@ end
 function git_object_hash(
     emit::Function,
     kind::AbstractString,
-    HashType::DataType,
-)
+    ::Type{HashType},
+) where HashType <: SHA.SHA_CTX
     ctx = HashType()
     body = codeunits(sprint(emit))
     SHA.update!(ctx, codeunits("$kind $(length(body))\0"))
@@ -265,9 +296,9 @@ end
 function git_file_hash(
     tar::IO,
     size::Integer,
-    HashType::DataType;
+    ::Type{HashType};
     buf::Vector{UInt8} = Vector{UInt8}(undef, DEFAULT_BUFFER_SIZE),
-)
+) where HashType <: SHA.SHA_CTX
     ctx = HashType()
     SHA.update!(ctx, codeunits("blob $size\0"))
     # TODO: this largely duplicates the logic of read_data
@@ -275,8 +306,8 @@ function git_file_hash(
     # where you write data to an IO object and it maintains a hash
     padded_size = round_up(size)
     while padded_size > 0
-        max_read_len = Int(min(padded_size, length(buf)))
-        read_len = readbytes!(tar, buf, max_read_len)
+        max_read_len = Int(min(padded_size, length(buf)))::Int
+        read_len = Int(readbytes!(tar, buf, max_read_len))::Int
         read_len < max_read_len && eof(tar) && throw(EOFError())
         nonpadded_view = view(buf, 1:Int(min(read_len, size)))
         SHA.update!(ctx, nonpadded_view)
@@ -304,8 +335,9 @@ function check_skeleton_header(
     buf::Vector{UInt8} = Vector{UInt8}(undef, DEFAULT_BUFFER_SIZE),
 )
     hdr = read_standard_header(skeleton, buf=buf)
-    hdr.type == :g && hdr.path == SKELETON_MAGIC ||
+    if hdr === nothing || !(hdr.type == :g && hdr.path == SKELETON_MAGIC)
         error("not a skeleton file: $skeleton")
+    end
     skip_data(skeleton, hdr.size)
 end
 
@@ -318,31 +350,62 @@ function read_tarball(
 )
     write_skeleton_header(skeleton, buf=buf)
     # symbols for path types except symlinks store the link
-    paths = Dict{String,Union{Symbol,String}}()
+    paths = Dict{String,Any}()
     globals = Dict{String,String}()
     while !eof(tar)
         hdr = read_header(tar, globals=globals, buf=buf, tee=skeleton)
         hdr === nothing && break
+        err = nothing
+        # normalize path and check for symlink attacks
+        path = ""
+        for part in split(hdr.path, '/')
+            # check_header checks for ".." later
+            (isempty(part) || part == ".") && continue
+            if err === nothing && get(paths, path, nothing) isa String
+                err = """
+                Tarball contains path with symlink prefix:
+                - path = $(repr(hdr.path))
+                - prefix = $(repr(path))
+                Refusing to extract — possible attack!
+                """
+            end
+            path = isempty(path) ? String(part) : "$path/$part"
+        end
+        hdr′ = Header(hdr, path=path)
+        # check that hardlinks refer to already-seen files
+        if err === nothing && hdr.type == :hardlink
+            parts = filter!(split(hdr.link, '/')) do part
+                # check_header checks for ".." later
+                !isempty(part) && part != "."
+            end
+            link = join(parts, '/')
+            hdr = Header(hdr, link=link)
+            hdr′ = Header(hdr′, link=link)
+            what = get(paths, link, Symbol("non-existent"))
+            if what isa Int64 # plain file
+                hdr′ = Header(hdr′, size=what)
+            else
+                err = """
+                Tarball contains hardlink with $what target:
+                - path = $(repr(hdr.path))
+                - target = $(repr(hdr.link))
+                Refusing to extract — possible attack!
+                """
+            end
+        end
         # check if we should extract or skip
-        if !predicate(hdr)
+        if !(predicate(hdr′)::Bool) # pass normalized header
             skip_data(tar, hdr.size)
             continue
         end
         check_header(hdr)
-        # normalize path and check for symlink attacks
-        path = ""
-        for part in split(hdr.path, '/')
-            (isempty(part) || part == ".") && continue
-            # check_header doesn't allow ".." in path
-            get(paths, path, nothing) isa String && error("""
-            Refusing to extract path with symlink prefix, possible attack
-             * path to extract: $(repr(hdr.path))
-             * symlink prefix: $(repr(path))
-            """)
-            isempty(path) || (paths[path] = :directory)
-            path = isempty(path) ? part : "$path/$part"
-        end
-        paths[path] = hdr.type == :symlink ? hdr.link : hdr.type
+        err === nothing || error(err)
+        # record info about path
+        paths[path] =
+            hdr.type == :symlink ? hdr.link :
+            hdr.type == :file    ? hdr.size :
+            hdr.type
+        # apply callback, checking that it consumes IO correctly
         before = applicable(position, tar) ? position(tar) : 0
         callback(hdr, split(path, '/', keepempty=false))
         applicable(position, tar) || continue
@@ -434,7 +497,7 @@ function read_extended_metadata(
         while j ≤ size
             byte = data[j += 1]
             byte == UInt8(' ') && break
-            UInt8('0') ≤ byte ≤ UInt8('9') || malformed()
+            UInt8('0') ≤ byte ≤ UInt8('9') || malformed()
             m, fm = mul_with_overflow(m, 10)
             m, fa = add_with_overflow(m, Int(byte - UInt8('0')))
             fm | fa &&
@@ -445,14 +508,14 @@ function read_extended_metadata(
             byte = data[k += 1]
             byte == UInt8('=') && break
         end
-        # data[i+1:j-1] is the length in decimal
-        # data[j+1:k-1] is the key string
-        # data[k+1:l-1] is the value string
-        # data[i] is end of previous
-        # data[j] is ` ` (space)
-        # data[k] is `=` (equals)
-        # data[l] is `\n` (newline)
-        i+1 < j < k < l || malformed()
+        # data[i]       is end of previous
+        # data[i+1:j-1] is length in decimal
+        # data[j]       is ` ` (space)
+        # data[j+1:k-1] is key string
+        # data[k]       is `=` (equals)
+        # data[k+1:l-1] is value string
+        # data[l]       is `\n` (newline)
+        i+1 < j < k < l || malformed()
         @assert data[j] == UInt8(' ')
         @assert data[k] == UInt8('=')
         data[l] == UInt('\n') || malformed()
@@ -524,7 +587,7 @@ function read_standard_header(
     # zero block indicates end of tarball
     if all(iszero, data)
         while !eof(io)
-            r = readbytes!(io, buf)
+            r = Int(readbytes!(io, buf))::Int
             write(tee, view(buf, 1:r))
         end
         return nothing
@@ -535,7 +598,8 @@ function read_standard_header(
         check_checksum_field(buf)
     catch err
         if err isa ErrorException
-            msg = match(r"^(.*?)\s*\[header block data\]"s, err.msg).captures[1]
+            m = match(r"^(.*?)\s*\[header block data\]"s, String(err.msg)::String)::RegexMatch
+            msg = something(m.captures[1])
             msg = "This does not appear to be a TAR file/stream — $msg. Note: Tar.jl does not handle decompression; if the tarball is compressed you must use an external command like `gzcat` or package like CodecZlib.jl to decompress it. See the README file for examples."
             err = ErrorException(msg)
         end
@@ -549,7 +613,7 @@ function read_standard_header(
     link = read_header_str(buf, :linkname)
     prefix = read_header_str(buf, :prefix)
     # check that mode isn't too big
-    mode ≤ typemax(typemax(UInt16)) ||
+    mode ≤ typemax(typemax(UInt16)) ||
         header_error(buf, "mode value too large: $(string(mode, base=8))")
     # combine prefix & name fields
     path = isempty(prefix) ? name : "$prefix/$name"
@@ -604,7 +668,7 @@ function read_header_int(buf::AbstractVector{UInt8}, fld::Symbol)
         before && byte == UInt8(' ') && continue
         byte in (0x00, UInt8(' ')) && break
         UInt8('0') <= byte <= UInt8('7') || header_error(buf, fld)
-        if leading_zeros(n) <= 3
+        if leading_zeros(n) <= 3
             val = String(buf[r])
             header_error(buf, "octal integer $fld value too large: $(repr(val))")
         end
@@ -622,7 +686,7 @@ function read_header_bin(
 )
     n = Int64(0)
     for i in r
-        if leading_zeros(n) <= 8
+        if leading_zeros(n) <= 8
             val = String(buf[r])
             header_error(buf, "binary integer $fld value too large: $(repr(val))")
         end
@@ -648,14 +712,17 @@ function read_data(
 )::Nothing
     padded_size = round_up(size)
     while padded_size > 0
-        max_read_len = Int(min(padded_size, length(buf)))
-        read_len = readbytes!(tar, buf, max_read_len)
+        max_read_len = Int(min(padded_size, length(buf)))::Int
+        read_len = Int(readbytes!(tar, buf, max_read_len))::Int
         write(tee, view(buf, 1:read_len))
         read_len < max_read_len && eof(tar) && throw(EOFError())
         size -= write(file, view(buf, 1:Int(min(read_len, size))))
         padded_size -= read_len
     end
-    @assert size == padded_size == 0
+    @assert size == padded_size == 0 """
+        size == padded_size == 0
+        This error may be a symptom of insufficient disk space on the device the tarball is being written to.
+        """
     return
 end
 
